/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import dk.vajhoej.record.ArrayField;
import dk.vajhoej.record.FieldType;
import dk.vajhoej.record.StructField;

/**
 * Class FieldInfo contains information about a ISAM key field needed for access.
 */
public class KeyInfo {
    private static Logger log = Logger.getLogger(KeyInfo.class.getName());
	private FieldType typ;
	private Field field;
	/**
	 * Create instance of KeyInfo.
	 * @param typ field type in native struct
	 * @param field field type in Java class
	 */
	public KeyInfo(FieldType typ, Field field) {
		this.typ = typ;
		this.field = field;
	}
	/**
	 * Get type.
	 * @return field type in native struct
	 */
	public FieldType getTyp() {
		return typ;
	}
	/**
	 * Get field
	 * @return field type in Java class
	 */
	public Field getField() {
		return field;
	}
	/**
	 * Analyze class.
	 * @param clz class to analyze
	 * @return KeyInfo array for class
	 */
    public static KeyInfo[] analyze(Class<?> clz) throws IsamException {
        List<KeyInfo> keys = new ArrayList<KeyInfo>();
        analyze(clz, keys);
        if(keys.size() == 0) {
            throw new IsamException("No keys defined for struct " + clz.getName());
        }
        log.finer(clz.getName() + " analyzed for KeyInfo array");
        return keys.toArray(new KeyInfo[keys.size()]);
    }
	private static void analyze(Class<?> clz, List<KeyInfo> keys) throws IsamException {
	    if(!clz.getSuperclass().equals(Object.class)) {
	        analyze(clz.getSuperclass(), keys);
	    }
	    int offset = keys.size();
		Field[] refl = clz.getDeclaredFields();
        for(int i = 0; i < refl.length; i++) {
            if((refl[i].getModifiers() & (Modifier.STATIC | Modifier.TRANSIENT)) == 0) {
	            KeyField kf = refl[i].getAnnotation(KeyField.class);
	            if(kf != null) {
	                keys.add(null);
	            }
		    }
		}
        for(int i = 0; i < refl.length; i++) {
            if((refl[i].getModifiers() & (Modifier.STATIC | Modifier.TRANSIENT)) == 0) {
                KeyField kf = refl[i].getAnnotation(KeyField.class);
                if(kf != null) {
                    StructField sf = refl[i].getAnnotation(StructField.class);
                    if(sf == null) {
                        throw new IsamException("Key is not a struct field for struct " + clz.getName());
                    }
                    ArrayField af = refl[i].getAnnotation(ArrayField.class);
                    if(af != null) {
                        throw new IsamException("Key is an array for struct " + clz.getName());
                    }
                    switch(sf.type()) {
                        case INT1:
                        case INT2:
                        case INT4:
                        case INT8:
                        case FIXSTR:
                        case FIXSTRNULTERM:
                        case VARSTR:
                        case VARFIXSTR:
                            // OK
                            break;
                        case UINT1:
                        case UINT2:
                        case UINT4:
                        case FP4:
                        case FP8:
                        case BOOLEAN:
                        case BIT:
                        case JAVATIME:
                        case UNIXTIME:
                        case VMSTIME:
                        case PACKEDBCD:
                        case ZONEDBCD:
                        case VAXFP4:
                        case VAXFP8:
                        case STRUCT:
                        default:
                            throw new IsamException("Key is not INT or STR for struct " + clz.getName());
                    }
                    if(kf.n() >= offset && kf.n() < keys.size()) {
                        if(keys.get(kf.n()) == null) {
                            refl[i].setAccessible(true);
                            keys.set(kf.n(), new KeyInfo(refl[i].getAnnotation(StructField.class).type(), refl[i]));
                        } else {
                            throw new IsamException("Duplicate key numbers for struct " + clz.getName());
                        }
                    } else {
                        throw new IsamException("Invalid key number for struct " + clz.getName());
                    }
                }
            }
        }
	}
}
