/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.remote;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Level;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.Key;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.StructReader;
import dk.vajhoej.record.StructWriter;

class ByteHelper {
    // String format:
    //   1 byte - length
    //   N bytes - UTF-8
    static void writeString(OutputStream os, String s) throws IOException {
        os.write(s.length());
        os.write(s.getBytes("UTF-8"));
    }
    static String readString(InputStream is) throws IOException {
        int len = is.read();
        byte[] b = new byte[len];
        is.read(b, 0, len);
        return new String(b, "UTF-8");
    }
    // Class format:
    //   String - class name
    static <T> void writeClass(OutputStream os, Class<T> t) throws IOException {
        writeString(os, t.getName());
    }
    static Class<?> readClass(InputStream is) throws ClassNotFoundException, IOException {
        return Class.forName(readString(is));
    }
    // IsamException format:
    //   String - exception message
    static void writeException(OutputStream os, IsamException e) throws IOException {
        writeString(os, e.getMessage());
    }
    static IsamException readException(InputStream is) throws IOException {
        return new IsamException(readString(is));
    }
    // Key format:
    //   1 byte - index
    //   Class - type
    //   String - value
    static <TK> void writeKey(OutputStream os, Key<TK> k) throws IOException {
        os.write(k.getIx());
        writeClass(os, k.getValue().getClass());
        writeString(os, k.getValue().toString());
    }
    @SuppressWarnings("unchecked")
    static Key<?> readKey(InputStream is) throws IOException, IllegalArgumentException, SecurityException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException {
        int ix = is.read();
        Class<?> clz = readClass(is);
        String strval = readString(is);
        Object val = clz.getConstructor(String.class).newInstance(strval);
        return new Key(ix, val);
    }
    // Object format:
    //   1 byte - 0=>value, 1=>null
    //   if value:
    //     Class - type
    //     2 bytes - length
    //     N bytes - record
    static <T> void writeObject(OutputStream os, Logger log, T o) throws IOException, RecordException {
        if(o == null) {
            os.write(1);
        } else {
            os.write(0);
            writeClass(os, o.getClass());
            StructWriter sw = new StructWriter();
            sw.write(o);
            byte[] b = sw.getBytes();
            log.finest("Length: " + b.length);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Bytes:" + LogHelper.byteArrayToString(b));
            }
            int len = b.length;
            os.write(len >> 8);
            os.write(len & 0xFF);
            os.write(b);
        }
    }
    static Object readObject(InputStream is, Logger log) throws IOException, RecordException, ClassNotFoundException {
        if(is.read() == 1) {
            return null;
        } else {
            Class<?> clz = readClass(is);
            int len = is.read();
            len = len << 8 | is.read();
            byte[] b = new byte[len];
            is.read(b, 0, len);
            log.finest("Length: " + b.length);
            if(log.isLoggable(Level.FINEST)) {
                log.finest("Bytes:" + LogHelper.byteArrayToString(b));
            }
            StructReader sr = new StructReader(b);
            return sr.read(clz);
        }
    }
}
