/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.remote;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.Key;
import dk.vajhoej.record.RecordException;

/**
 * Interface IsamProtocolClient defines the API for remote ISAM sources clients.
 * This interface is not intended for use by applications.
 */
public interface IsamProtocolClient extends IsamProtocol {
    /**
     * Send preferred version.
     * @throws IsamException if problem with ISAM source
     */
    public void handshake() throws IsamException;
    /**
     * Read record from ISAM source.
     * @param t type to read
     * @param k key of record to read
     * @return object read
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T, TK> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException;
    /**
     * Create record in ISAM source.
     * @param o object to create
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T> void create(T o) throws IsamException, RecordException;
    /**
     * Update record in ISAM source.
     * @param o object to update
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T> void update(T o) throws IsamException, RecordException;
    /**
     * Delete record from ISAM source.
     * @param t type to delete
     * @param k key of record to delete
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T, TK> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException;
    /**
     * Close ISAM source.
     * @throws IsamException if problem with ISAM source
     */
    public void close() throws IsamException;
    /**
     * Find all records matching criteria.
     * @param t type to find
     * @param k key to find
     * @return result
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T, TK> IsamResult<T> readXX(int cmd, Class<T> t, Key<TK> k) throws IsamException, RecordException;
    /**
     * Read record.
     * @return true if record read false if record not read
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public boolean resultRead() throws IsamException, RecordException;
    /**
     * Return current record.
     * @return current record
     * @throws IsamException if problem with ISAM source
     * @throws RecordException if problem converting data
     */
    public <T> T resultCurrent() throws IsamException, RecordException;
    /**
     * Close.
     * @throws IsamException if problem with ISAM source
     */
    public void resultClose() throws IsamException;
}
