/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.rms;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Parse C source code and generate Java bean classes for types.
 * Run as:
 * <pre>
 *   java dk.vajhoej.vms.rms.CParse <c or h file>
 * </pre>
 */
public class CParse {
    private static final int MISSING = -1;
    private static Map<String, String> c2java = new HashMap<String, String>(); 
    private static Map<String, String> c2struct = new HashMap<String, String>();
    private static Map<String, Integer> c2length = new HashMap<String, Integer>();
    private static void map(String c, String java, String struct, int length) {
        c2java.put(c, java);
        c2struct.put(c, struct);
        c2length.put(c, length);
    }
    static {
        map("char", "byte", "INT1", MISSING);
        map("signed char", "byte", "INT1", MISSING);
        map("unsigned char", "short", "UINT1", MISSING);
        map("int", "int", "INT4", MISSING);
        map("short", "short", "INT2", MISSING);
        map("short int", "short", "INT2", MISSING);
        map("long", "int", "INT4", MISSING);
        map("long int", "int", "INT4", MISSING);
        map("signed", "int", "INT4", MISSING);
        map("signed int", "int", "INT4", MISSING);
        map("signed short", "short", "INT2", MISSING);
        map("signed short int", "short", "INT2", MISSING);
        map("signed long", "int", "INT4", MISSING);
        map("signed long int", "int", "INT4", MISSING);
        map("unsigned", "long", "UINT4", MISSING);
        map("unsigned int", "long", "UINT4", MISSING);
        map("unsigned short", "int", "UINT2", MISSING);
        map("unsigned short int", "int", "UINT2", MISSING);
        map("unsigned long", "long", "UINT4", MISSING);
        map("unsigned long int", "long", "UINT4", MISSING);
        map("long long", "long", "INT8", MISSING);
        map("long long int", "long", "INT8", MISSING);
        map("float", "double", "FP4", MISSING);
        map("double", "double", "FP8", MISSING);
    }
    private static class Field {
        private String name;
        private String javaType;
        private String structType;
        private int length;
        private int key;
        public Field(String name, String type, int key) {
            this.name = name.toLowerCase();
            this.javaType = c2java.get(type);
            this.structType = c2struct.get(type);
            this.length = c2length.get(type);
            this.key = key;
        }
        public String getName() {
            return name;
        }
        public String getJavaType() {
            return javaType;
        }
        public String getStructType() {
            return structType;
        }
        public int getLength() {
            return length;
        }
        public int getKey() {
            return key;
        }
    }
    private static String capitalize(String s) {
        return s.substring(0, 1).toUpperCase() + s.substring(1);
    }
    private static String stripComments(String s) {
        return s.replaceAll("/\\*.*?\\*/", "");
    }
    private static void genClass(String recnam, List<Field> fld) {
        System.out.println("import java.io.Serializable;");
        System.out.println();
        System.out.println("import dk.vajhoej.isam.KeyField;");
        System.out.println("import dk.vajhoej.record.Alignment;");
        System.out.println("import dk.vajhoej.record.FieldType;");
        System.out.println("import dk.vajhoej.record.Struct;");
        System.out.println("import dk.vajhoej.record.StructField;");
        System.out.println();
        System.out.println("@Struct(alignment=Alignment.NATURAL,endpad=true)");
        System.out.println("public class " + recnam + " implements Serializable {");
        System.out.println("    private static final long serialVersionUID = 1L;");
        int nfld = 0;
        for(Field fnt : fld) {
            if(fnt.getKey() > MISSING) {
                System.out.println("    @KeyField(n=" + fnt.getKey() + ")");
            }
            if(fnt.getLength() > MISSING) {
                System.out.println("    @StructField(n=" + nfld + ",type=FieldType." + fnt.getStructType() + ",length=" + fnt.getLength() + ")");
            } else {
                System.out.println("    @StructField(n=" + nfld + ",type=FieldType." + fnt.getStructType() + ")");
            }
            System.out.println("    private " + fnt.getJavaType() + " " + fnt.getName() + ";");
            nfld++;
        }
        for(Field fnt : fld) {
            System.out.println("    public " + fnt.getJavaType() + " get" + capitalize(fnt.getName()) + "() {");
            System.out.println("        return " + fnt.getName() + ";");
            System.out.println("    }");
            System.out.println("    public void set" + capitalize(fnt.getName()) + "(" + fnt.getJavaType() + " " + fnt.getName() + ") {");
            System.out.println("        this." + fnt.getName() + " = " + fnt.getName() + ";");
            System.out.println("    }");
        }
        System.out.println("}");
        System.out.println();
    }
    private static Pattern chrarrpat = Pattern.compile("(?:typedef\\s+char\\s+)(\\w+)(?:\\s*\\[)(\\d+)(?:\\];)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    private static void processStrings(String s) {
        Matcher chrarr = chrarrpat.matcher(s);
        while(chrarr.find()) {
            map(chrarr.group(1), "String", "FIXSTRNULTERM", Integer.parseInt(chrarr.group(2)));
        }
    }
    private static Pattern fldpat = Pattern.compile("((\\w+)(\\s+\\w+)*)(?:\\s+)(\\w+)(((?:\\s*:\\s*)(\\d+))|((?:\\[)(\\d+)(?:\\]))|())(?:\\s*;)");
    private static void processRecord(String recnam, String s) {
        List<Field> fld = new ArrayList<Field>();
        Matcher m = fldpat.matcher(s);
        boolean first = true;
        while(m.find()) {
            if(m.group(5).equals("")) {
                fld.add(new Field(m.group(4), m.group(1), first ? 0 : MISSING));
            } else {
                if(m.group(6) != null && !m.group(6).equals("")) {
                    int len = Integer.parseInt(m.group(7));
                    String typnam = "__bit" + len;
                    map(typnam, "int", "BIT", len);
                    fld.add(new Field(m.group(4), typnam, first ? 0 : MISSING));
                } else if(m.group(8) != null && !m.group(8).equals("")) {
                    int len = Integer.parseInt(m.group(9));
                    String typnam = "__cstr" + len;
                    map(typnam, "String", "FIXSTRNULTERM", len);
                    fld.add(new Field(m.group(4), typnam, first ? 0 : MISSING));
                }
            }
            first = false;
        }
        genClass(capitalize(recnam.toLowerCase()), fld);
    }
    private static Pattern recpat = Pattern.compile("(?:struct\\s+)(\\w+)(?:\\s+\\{)(.*?)(?:\\s+}\\s*;)", Pattern.CASE_INSENSITIVE + Pattern.DOTALL);
    /**
     * Process a file.
     * @param s content of file
     */
    public static void processFile(String s) {
        String s2 = stripComments(s);
        processStrings(s2);
        Matcher m = recpat.matcher(s2);
        while(m.find()) {
            processRecord(m.group(1), m.group(2));
            map(m.group(1), capitalize(m.group(1)), "STRUCT", MISSING);
        }
    }
    /**
     * PascalParse main.
     * @param args command line arguments - first and only argument is name of pascal file 
     * @throws IOException if problem with file
     */
    public static void main(String[] args) {
        try {
            BufferedReader br = new BufferedReader(new FileReader(args[0]));
            StringBuilder sb = new StringBuilder();
            String line;
            while((line = br.readLine()) != null) {
                sb.append(line);
                sb.append(System.getProperty("line.separator"));
            }
            br.close();
            processFile(sb.toString());
        } catch(Exception ex) {
            ex.printStackTrace();
        }
    }
}
